"""add source metadata column

Revision ID: f880962b9187
Revises: 1f8b94fca61b
Create Date: 2025-01-03 15:03:09.785169

"""

from typing import Sequence, Union

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision: str = "f880962b9187"
down_revision: Union[str, None] = "1f8b94fca61b"
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column("memory", sa.Column("source_metadata", sa.Text(), nullable=True))
    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column("memory", "source_metadata")
