"""add system instruct to user preference

Revision ID: 71ffabe55f21
Revises: b360a1f1b06e
Create Date: 2024-12-13 21:15:57.911780

"""

from typing import Sequence, Union

import sqlalchemy as sa
from alembic import op
from sqlmodel.sql.sqltypes import AutoString

# revision identifiers, used by Alembic.
revision: str = "71ffabe55f21"
down_revision: Union[str, None] = "b360a1f1b06e"
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column("userpreference", sa.Column("system_persona", AutoString(), nullable=True))
    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column("userpreference", "system_persona")
    # ### end Alembic commands ###
