"""add source metadata column

Revision ID: 0fe803cc98f1
Revises: 7781f460cde2
Create Date: 2025-01-03 15:03:21.787143

"""

from typing import Sequence, Union

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision: str = "0fe803cc98f1"
down_revision: Union[str, None] = "7781f460cde2"
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column("memory", sa.Column("source_metadata", sa.Text(), nullable=True))
    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column("memory", "source_metadata")
    # ### end Alembic commands ###
