"""add assistant name

Revision ID: 66a25cb7d3ef
Revises: 4ed800abf24a
Create Date: 2024-12-26 17:07:33.257863

"""

from typing import Sequence, Union

import sqlalchemy as sa
from alembic import op
from sqlmodel.sql.sqltypes import AutoString

# revision identifiers, used by Alembic.
revision: str = "66a25cb7d3ef"
down_revision: Union[str, None] = "4ed800abf24a"
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###

    op.add_column("userpreference", sa.Column("assistant_name", AutoString(), nullable=True))
    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column("userpreference", "assistant_name")
    # ### end Alembic commands ###
