"""rm vector timestamps

Revision ID: 4ed800abf24a
Revises: 892ce7b9a77b
Create Date: 2024-12-24 20:22:49.843503

"""

from typing import Sequence, Union

import sqlalchemy as sa
from alembic import op
from sqlalchemy.dialects import postgresql

# revision identifiers, used by Alembic.
revision: str = "4ed800abf24a"
down_revision: Union[str, None] = "892ce7b9a77b"
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column("vectorstorage", "updated_at")
    op.drop_column("vectorstorage", "created_at")
    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column("vectorstorage", sa.Column("created_at", postgresql.TIMESTAMP(), autoincrement=False, nullable=False))
    op.add_column("vectorstorage", sa.Column("updated_at", postgresql.TIMESTAMP(), autoincrement=False, nullable=False))
    # ### end Alembic commands ###
